from django.db import models
from edc_action_item.managers import (
    ActionIdentifierModelManager,
    ActionIdentifierSiteManager,
)
from edc_action_item.models import ActionModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_pdf_reports.model_mixins import PdfReportModelMixin
from edc_sites.model_mixins import SiteModelMixin

from ...constants import AE_INITIAL_ACTION
from ...pdf_reports import AePdfReport
from .ae_initial_ae_model_mixin import AeInitialAeModelMixin
from .ae_initial_fields_model_mixin import AeInitialFieldsModelMixin
from .ae_initial_methods_model_mixin import AeInitialMethodsModelMixin
from .ae_initial_sae_model_mixin import AeInitialSaeModelMixin
from .ae_initial_susar_model_mixin import AeInitialSusarModelMixin
from .ae_initial_tmg_model_mixin import AeInitialTmgModelMixin


class AeInitialModelMixin(
    SiteModelMixin,
    ActionModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    AeInitialFieldsModelMixin,
    AeInitialMethodsModelMixin,
    AeInitialAeModelMixin,
    AeInitialSaeModelMixin,
    AeInitialSusarModelMixin,
    AeInitialTmgModelMixin,
    PdfReportModelMixin,
    models.Model,
):
    action_name = AE_INITIAL_ACTION

    pdf_report_cls = AePdfReport

    objects = ActionIdentifierModelManager()

    on_site = ActionIdentifierSiteManager()

    class Meta(NonUniqueSubjectIdentifierFieldMixin.Meta):
        abstract = True
        verbose_name = "AE Initial Report"
        indexes = (
            NonUniqueSubjectIdentifierFieldMixin.Meta.indexes
            + ActionModelMixin.Meta.indexes
            + [models.Index(fields=["subject_identifier", "action_identifier", "site"])]
        )



from django.db import models
from edc_action_item.managers import (
    ActionIdentifierModelManager,
    ActionIdentifierSiteManager,
)
from edc_action_item.models import ActionModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import ReportStatusModelMixin
from edc_search.model_mixins import SearchSlugModelMixin
from edc_sites.model_mixins import SiteModelMixin

from ...constants import AE_TMG_ACTION
from .ae_tmg_fields_model_mixin import AeTmgFieldsModelMixin
from .ae_tmg_methods_model_mixin import AeTmgMethodsModelMixin


class AeTmgModelMixin(
    AeTmgMethodsModelMixin,
    SiteModelMixin,
    ActionModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    AeTmgFieldsModelMixin,
    ReportStatusModelMixin,
    SearchSlugModelMixin,
    models.Model,
):
    action_name = AE_TMG_ACTION

    objects = ActionIdentifierModelManager()

    on_site = ActionIdentifierSiteManager()

    class Meta(NonUniqueSubjectIdentifierFieldMixin.Meta, ActionModelMixin.Meta):
        abstract = True
        verbose_name = "AE TMG Report"
        verbose_name_plural = "AE TMG Reports"
        indexes = (
            NonUniqueSubjectIdentifierFieldMixin.Meta.indexes
            + ActionModelMixin.Meta.indexes
            + [models.Index(fields=["subject_identifier", "action_identifier", "site", "id"])]
        )



from django.db import models
from edc_action_item.managers import (
    ActionIdentifierModelManager,
    ActionIdentifierSiteManager,
)
from edc_action_item.models import ActionModelMixin
from edc_identifier.model_mixins import UniqueSubjectIdentifierFieldMixin
from edc_model.validators import date_not_future, datetime_not_future
from edc_protocol.validators import datetime_not_before_study_start
from edc_sites.model_mixins import SiteModelMixin
from edc_utils import get_utcnow

from ...constants import DEATH_REPORT_ACTION


class SimpleDeathReportModelMixin(
    UniqueSubjectIdentifierFieldMixin,
    SiteModelMixin,
    ActionModelMixin,
    models.Model,
):
    action_name = DEATH_REPORT_ACTION

    death_date_field = "death_date"

    report_datetime = models.DateTimeField(
        verbose_name="Report Date",
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=get_utcnow,
    )

    death_datetime = models.DateTimeField(
        validators=[datetime_not_future],
        verbose_name="Date and Time of Death",
        null=True,
        blank=False,
    )

    death_date = models.DateField(
        validators=[date_not_future],
        verbose_name="Date of Death",
        null=True,
        blank=False,
    )

    narrative = models.TextField(verbose_name="Narrative")

    objects = ActionIdentifierModelManager()

    on_site = ActionIdentifierSiteManager()

    def natural_key(self):
        return (self.action_identifier,)

    class Meta(ActionModelMixin.Meta):
        abstract = True
        verbose_name = "Death Report"
        verbose_name_plural = "Death Reports"
        indexes = ActionModelMixin.Meta.indexes + [
            models.Index(fields=["subject_identifier", "action_identifier", "site", "id"])
        ]



from django.db import models


class AeTmgMethodsModelMixin(models.Model):
    def __str__(self):
        return f"{self.action_identifier[-9:]}"

    def save(self, *args, **kwargs):
        self.subject_identifier = self.ae_initial.subject_identifier
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.action_identifier,)  # noqa

    def get_action_item_reason(self):
        return self.ae_initial.ae_description

    def get_search_slug_fields(self):
        fields = super().get_search_slug_fields()
        fields.append("subject_identifier")
        fields.append("report_status")
        return fields

    class Meta:
        abstract = True



from __future__ import annotations

from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple
from edc_action_item import action_fieldset_tuple
from edc_action_item.modeladmin_mixins import ActionItemModelAdminMixin
from edc_constants.constants import OTHER
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin

from ..utils import get_ae_model


class DeathReportTmgModelAdminMixin(
    ModelAdminSubjectDashboardMixin, ActionItemModelAdminMixin
):
    add_form_template: str = "edc_adverse_event/admin/change_form.html"
    change_list_template = "edc_adverse_event/admin/change_list.html"
    change_form_template = "edc_adverse_event/admin/change_form.html"

    fieldsets = (
        (None, {"fields": ("subject_identifier", "death_report", "report_datetime")}),
        (
            "Opinion of TMG",
            {
                "fields": (
                    "cause_of_death",
                    "cause_of_death_other",
                    "cause_of_death_agreed",
                    "narrative",
                    "report_status",
                    "report_closed_datetime",
                )
            },
        ),
        action_fieldset_tuple,
        audit_fieldset_tuple,
    )

    radio_fields = {
        "cause_of_death": admin.VERTICAL,
        "cause_of_death_agreed": admin.VERTICAL,
        "report_status": admin.VERTICAL,
    }

    search_fields = (
        "subject_identifier",
        "action_identifier",
        "death_report__action_identifier",
    )

    def get_list_display(self, request) -> tuple[str]:
        list_display = super().get_list_display(request)
        custom_fields = (
            "subject_identifier",
            "dashboard",
            "report_datetime",
            "cause",
            "agreed",
            "status",
            "report_closed_datetime",
        )
        return custom_fields + tuple(f for f in list_display if f not in custom_fields)

    def get_list_filter(self, request) -> tuple[str]:
        list_filter = super().get_list_filter(request)
        custom_fields = (
            "report_datetime",
            "report_status",
            "cause_of_death_agreed",
            "cause_of_death",
        )
        return custom_fields + tuple(f for f in list_filter if f not in custom_fields)

    def get_readonly_fields(self, request, obj=None) -> tuple[str]:
        fields = super().get_readonly_fields(request, obj)
        if obj:
            fields = fields + ("death_report",)
        return fields

    @staticmethod
    def status(obj=None):
        return obj.report_status.title()

    def cause(self, obj=None):
        if obj.cause_of_death.name == OTHER:
            return f"Other: {obj.cause_of_death_other}"
        return obj.cause_of_death

    cause.short_description = "Cause (TMG Opinion)"

    @staticmethod
    def agreed(obj=None):
        return obj.cause_of_death_agreed

    @property
    def death_report_model_cls(self):
        return get_ae_model("deathreport")

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "death_report":
            if request.GET.get("death_report"):
                kwargs["queryset"] = self.death_report_model_cls.objects.filter(
                    id__exact=request.GET.get("death_report", 0)
                )
            else:
                kwargs["queryset"] = self.death_report_model_cls.objects.none()
        return super().formfield_for_foreignkey(db_field, request, **kwargs)



from django.db import models
from edc_constants.choices import YES_NO, YES_NO_NA
from edc_constants.constants import NOT_APPLICABLE


class AeInitialSusarModelMixin(models.Model):

    """This model mixin is for the AE Initial."""

    susar = models.CharField(
        verbose_name="Is this a Suspected Unexpected Serious Adverse Reaction (SUSAR)?",
        choices=YES_NO,
        max_length=5,
        help_text=(
            "If yes, SUSAR must be reported to Principal " "Investigator and TMG immediately,"
        ),
    )

    susar_reported = models.CharField(
        verbose_name="Is SUSAR reported?",
        max_length=5,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    class Meta:
        abstract = True



from django import forms

from ..modelform_mixins import DeathReportTmgSecondModelFormMixin
from ..utils import get_ae_model


class DeathReportTmgSecondForm(DeathReportTmgSecondModelFormMixin, forms.ModelForm):
    class Meta(DeathReportTmgSecondModelFormMixin.Meta):
        model = get_ae_model("deathreporttmgsecond")
        fields = "__all__"



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date, datetime
from typing import TYPE_CHECKING, Type

from django.conf import settings
from django.urls import reverse
from edc_model.models import BaseUuidModel
from edc_utils import convert_php_dateformat

if TYPE_CHECKING:
    from django.contrib import admin

    from .modeladmin_mixins import AdverseEventModelAdminMixin

    class ModelAdmin(AdverseEventModelAdminMixin, admin.ModelAdmin):
        ...

    class Model(BaseUuidModel):
        report_datetime: datetime
        ...


@dataclass(order=True)
class ColumnItem:
    """Prepare a label and url for use in list_display.

    Used my AE modeladmin"""

    modeladmin: ModelAdmin = field(compare=False)
    obj: Model = field(compare=False)
    search: str = field(compare=False)
    date_field: str | None = field(default="report_datetime", compare=False)
    verbose_name: str = field(init=False, repr=False, compare=False)
    formatted_date: str = field(init=False, repr=False, compare=False)
    model_cls: Type[Model] = field(init=False, repr=False, compare=False)
    date_value: date = field(init=False, repr=False, compare=True)

    def __post_init__(self):
        try:
            self.date_value = getattr(self.obj, self.date_field).date()
        except AttributeError:
            self.date_value = getattr(self.obj, self.date_field)
        self.verbose_name = self.obj._meta.verbose_name
        self.formatted_date = self.date_value.strftime(
            convert_php_dateformat(settings.SHORT_DATE_FORMAT)
        )
        self.model_cls = self.obj.__class__

    @property
    def anchor(self) -> str:
        return (
            f'<a title="go to {self.verbose_name} for '
            f'{self.formatted_date}" href="{self.url}?q={self.search}">'
            f"<span nowrap>{self.verbose_name}</span></a>"
        )

    @property
    def url(self) -> str:
        url_name = "_".join(self.obj._meta.label_lower.split("."))
        namespace = self.modeladmin.admin_site.name
        return reverse(f"{namespace}:{url_name}_changelist")



from django import forms

from ..modelform_mixins import DeathReportTmgModelFormMixin
from ..utils import get_ae_model


class DeathReportTmgForm(DeathReportTmgModelFormMixin, forms.ModelForm):
    class Meta(DeathReportTmgModelFormMixin.Meta):
        model = get_ae_model("deathreporttmg")
        fields = "__all__"



from django.core import mail
from django.test import TestCase, override_settings
from edc_constants.constants import NO, YES
from edc_facility.import_holidays import import_holidays
from edc_list_data.site_list_data import site_list_data
from edc_reportable.constants import GRADE3, GRADE4, GRADE5
from model_bakery import baker

from adverse_event_app import list_data

from ...action_items import (
    AeFollowupAction,
    AeInitialAction,
    AeSusarAction,
    AeTmgAction,
    DeathReportAction,
    DeathReportTmgAction,
)
from ...notifications import AeInitialG3EventNotification, AeInitialG4EventNotification
from .mixins import DeathReportTestMixin


@override_settings(EDC_LIST_DATA_ENABLE_AUTODISCOVER=False)
class TestNotifications(DeathReportTestMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        site_list_data.initialize()
        site_list_data.register(list_data, app_name="adverse_event_app")
        site_list_data.load_data()
        import_holidays()

    def test_notifies_initial_ae_g3_not_sae(self):
        baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            ae_grade=GRADE3,
            sae=NO,
        )

        self.assertEqual(len(mail.outbox), 3)

        # AeInitial Action notification
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeInitialAction.notification_display_name in m.__dict__.get("subject")
                ]
            ),
        )

        # AeInitialG3EventNotification
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeInitialG3EventNotification.display_name in m.__dict__.get("subject")
                ]
            ),
        )

        # AeFollowupAction
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeFollowupAction.notification_display_name in m.__dict__.get("subject")
                ]
            ),
        )

    def test_notifies_initial_ae_g3_is_sae(self):
        baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            ae_grade=GRADE3,
            sae=YES,
        )

        self.assertEqual(len(mail.outbox), 4)

        # AeInitial Action notification
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeInitialAction.notification_display_name in m.__dict__.get("subject")
                ]
            ),
        )

        # AeInitialG3EventNotification
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeInitialG3EventNotification.display_name in m.__dict__.get("subject")
                ]
            ),
        )

        # AeFollowupAction
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeFollowupAction.notification_display_name in m.__dict__.get("subject")
                ]
            ),
        )

        # AeTmgAction
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeTmgAction.notification_display_name in m.__dict__.get("subject")
                ]
            ),
        )

    def test_notifies_initial_ae_g4_is_sae(self):
        baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            ae_grade=GRADE4,
            sae=YES,
        )

        self.assertEqual(len(mail.outbox), 4)

        # AeInitialG4EventNotification
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeInitialG4EventNotification.display_name in m.__dict__.get("subject")
                ]
            ),
        )

    def test_notifies_initial_ae_death(self):
        baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            ae_grade=GRADE5,
            sae=YES,
        )

        self.assertEqual(len(mail.outbox), 3)

        # AeInitial Action notification
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeInitialAction.notification_display_name in m.__dict__.get("subject")
                ]
            ),
        )

        # DeathReportAction
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if DeathReportAction.notification_display_name in m.__dict__.get("subject")
                ]
            ),
        )

        # AeTmgAction
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeTmgAction.notification_display_name in m.__dict__.get("subject")
                ]
            ),
        )

    def test_notifies_initial_ae_death_with_tmg(self):
        self.get_death_report()

        self.assertEqual(len(mail.outbox), 7)

        # AeInitial Action notification
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeInitialAction.notification_display_name in m.__dict__.get("subject")
                ]
            ),
        )

        # DeathReportAction
        self.assertIn(
            DeathReportAction.notification_display_name,
            "|".join([m.__dict__.get("subject") for m in mail.outbox]),
        )

        # DeathReportTmgAction
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if DeathReportTmgAction.notification_display_name
                    in m.__dict__.get("subject")
                ]
            ),
        )

    def test_notifies_initial_ae_susar(self):
        baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            ae_grade=GRADE4,
            sae=YES,
            susar=YES,
            susar_reported=NO,
        )
        self.assertEqual(len(mail.outbox), 5)

        # AeSusarAction
        self.assertEqual(
            1,
            len(
                [
                    m.__dict__.get("subject")
                    for m in mail.outbox
                    if AeSusarAction.notification_display_name in m.__dict__.get("subject")
                ]
            ),
        )



from typing import Any

from django.contrib.sites.shortcuts import get_current_site
from django.db.models.aggregates import Count
from django.urls import reverse
from django.views.generic import TemplateView
from edc_action_item.models.action_item import ActionItem
from edc_constants.constants import CLOSED, NEW, OPEN
from edc_dashboard.utils import get_bootstrap_version
from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin

from ...constants import AE_TMG_ACTION
from ...utils import get_adverse_event_admin_site, get_adverse_event_app_label


class TmgHomeView(EdcViewMixin, NavbarViewMixin, TemplateView):
    template_name = f"edc_adverse_event/bootstrap{get_bootstrap_version()}/tmg/tmg_home.html"
    navbar_selected_item = "tmg_home"

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        death_report_changelist_url = reverse(
            f"{get_adverse_event_admin_site()}:{get_adverse_event_app_label()}_"
            f"deathreporttmg_changelist"
        )

        # summarize closed reports by site
        summary = (
            ActionItem.objects.filter(action_type__name=AE_TMG_ACTION, status=CLOSED)
            .values("site__name")
            .annotate(count=Count("status"))
            .order_by("site__name")
        )
        # summarize new and open for notice
        qs = (
            ActionItem.objects.filter(action_type__name=AE_TMG_ACTION, status__in=[NEW, OPEN])
            .exclude(site__name=get_current_site(request=self.request).name)
            .values("status", "site__name")
            .annotate(items=Count("status"))
        )
        notices = []
        for item in qs.order_by("status", "site__name"):
            notices.append([item.get("site__name"), item.get("status"), item.get("items")])
        new_count = ActionItem.objects.filter(
            action_type__name=AE_TMG_ACTION,
            site__name=get_current_site(request=self.request).name,
            status=NEW,
        ).count()
        open_count = ActionItem.objects.filter(
            action_type__name=AE_TMG_ACTION,
            site__name=get_current_site(request=self.request).name,
            status=OPEN,
        ).count()
        closed_count = ActionItem.objects.filter(
            action_type__name=AE_TMG_ACTION,
            site__name=get_current_site(request=self.request).name,
            status=CLOSED,
        ).count()
        total_count = ActionItem.objects.filter(
            action_type__name=AE_TMG_ACTION,
            site__name=get_current_site(request=self.request).name,
            status__in=[NEW, OPEN, CLOSED],
        ).count()
        kwargs.update(
            {
                "new_count": new_count,
                "open_count": open_count,
                "closed_count": closed_count,
                "total_count": total_count,
                "summary": summary,
                "notices": notices,
                "death_report_changelist_url": death_report_changelist_url,
            }
        )
        return super().get_context_data(**kwargs)



from django.contrib import admin
from edc_model_admin.history import SimpleHistoryAdmin

from edc_adverse_event.modeladmin_mixins import (
    AeFollowupModelAdminMixin,
    AeInitialModelAdminMixin,
)

from .models import AeFollowup, AeInitial


@admin.register(AeInitial)
class AeInitialAdmin(AeInitialModelAdminMixin, SimpleHistoryAdmin):
    pass


@admin.register(AeFollowup)
class AeFollowupAdmin(AeFollowupModelAdminMixin, SimpleHistoryAdmin):
    pass



from django.apps import apps as django_apps
from django.db import models


class AeInitialMethodsModelMixin(models.Model):
    def __str__(self):
        return f"{self.action_identifier[-9:]} Grade {self.ae_grade}"

    def natural_key(self):
        return (self.action_identifier,)

    def get_action_item_reason(self):
        return self.ae_description

    @property
    def ae_follow_up_model_cls(self):
        return django_apps.get_model(f"{self._meta.app_label}.aefollowup")

    @property
    def ae_follow_ups(self):
        return self.ae_follow_up_model_cls.objects.filter(ae_initial=self).order_by(
            "report_datetime"
        )

    @property
    def description(self):
        """Returns a description."""
        return f"{self.action_identifier[-9:]} Grade-{self.ae_grade}. {self.ae_description}"

    class Meta:
        abstract = True



from __future__ import annotations

from random import choice

from edc_action_item.models import ActionItem
from edc_constants.constants import OTHER, YES
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from model_bakery import baker

from adverse_event_app.visit_schedules import visit_schedule

from ...models import CauseOfDeath


class DeathReportTestMixin:
    def setUp(self):
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)
        subject_consent = baker.make_recipe(
            "adverse_event_app.subjectconsent", subject_identifier="1234567"
        )
        self.subject_identifier = subject_consent.subject_identifier

        # put subject on schedule
        _, schedule = site_visit_schedules.get_by_onschedule_model(
            "adverse_event_app.onschedule"
        )
        schedule.put_on_schedule(
            subject_identifier=subject_consent.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )

    def get_death_report(
        self,
        cause_of_death: str | None = None,
        cause_of_death_other: str | None = None,
    ):
        causes_qs = CauseOfDeath.objects.exclude(name=OTHER)
        cause_of_death = (
            cause_of_death
            or causes_qs[choice([x for x in range(0, len(causes_qs))])]  # nosec B311
        )

        # create ae initial
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            susar=YES,
            susar_reported=YES,
            ae_grade=5,
            user_created="erikvw",
        )

        action_item = ActionItem.objects.get(
            subject_identifier=self.subject_identifier,
            parent_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.deathreport",
        )

        # create death report
        death_report = baker.make_recipe(
            "adverse_event_app.deathreport",
            subject_identifier=self.subject_identifier,
            action_identifier=action_item.action_identifier,
            cause_of_death=cause_of_death,
            cause_of_death_other=cause_of_death_other,
            user_created="erikvw",
        )
        return death_report



from django.db import models
from django.db.models.deletion import PROTECT
from edc_action_item.models.action_model_mixin import ActionModelMixin
from edc_consent.field_mixins.identity_fields_mixin import IdentityFieldsMixin
from edc_consent.field_mixins.personal_fields_mixin import PersonalFieldsMixin
from edc_consent.model_mixins import ConsentModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierModelMixin
from edc_model.models import BaseUuidModel
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_sites.model_mixins import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_schedule.model_mixins import OnScheduleModelMixin
from edc_visit_schedule.model_mixins.off_schedule_model_mixin import (
    OffScheduleModelMixin,
)
from edc_visit_tracking.model_mixins import VisitModelMixin

from edc_adverse_event.constants import STUDY_TERMINATION_CONCLUSION_ACTION
from edc_adverse_event.model_mixins import (
    AeFollowupModelMixin,
    AeInitialModelMixin,
    AesiModelMixin,
    AeSusarModelMixin,
    AeTmgModelMixin,
    DeathReportModelMixin,
    DeathReportTmgModelMixin,
    DeathReportTmgSecondModelMixin,
)


class SubjectConsent(
    SiteModelMixin,
    ConsentModelMixin,
    PersonalFieldsMixin,
    IdentityFieldsMixin,
    NonUniqueSubjectIdentifierModelMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    BaseUuidModel,
):
    class Meta(ConsentModelMixin.Meta):
        pass


class SubjectVisit(SiteModelMixin, VisitModelMixin, BaseUuidModel):
    pass


class CrfOne(SiteModelMixin, NonUniqueSubjectIdentifierModelMixin, BaseUuidModel):
    report_datetime = models.DateTimeField(default=get_utcnow)


class OnSchedule(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):

    """A model used by the system. Auto-completed by subject_consent."""

    def put_on_schedule(self):
        pass

    class Meta(OnScheduleModelMixin.Meta):
        pass


class StudyTerminationConclusion(
    SiteModelMixin, ActionModelMixin, OffScheduleModelMixin, BaseUuidModel
):
    action_name = STUDY_TERMINATION_CONCLUSION_ACTION

    subject_identifier = models.CharField(max_length=50, unique=True)

    class Meta(OffScheduleModelMixin.Meta):
        pass


class AeInitial(AeInitialModelMixin, BaseUuidModel):
    class Meta(AeInitialModelMixin.Meta):
        pass


class AeFollowup(AeFollowupModelMixin, BaseUuidModel):
    ae_initial = models.ForeignKey(AeInitial, on_delete=PROTECT)

    class Meta(AeFollowupModelMixin.Meta):
        pass


class Aesi(AesiModelMixin, BaseUuidModel):
    ae_initial = models.ForeignKey(AeInitial, on_delete=PROTECT)

    class Meta(AesiModelMixin.Meta):
        pass


class AeSusar(AeSusarModelMixin, BaseUuidModel):
    ae_initial = models.ForeignKey(AeInitial, on_delete=PROTECT)

    class Meta(AeSusarModelMixin.Meta):
        pass


class AeTmg(AeTmgModelMixin, BaseUuidModel):
    ae_initial = models.ForeignKey(AeInitial, on_delete=PROTECT)

    class Meta(AeTmgModelMixin.Meta):
        pass


class DeathReport(DeathReportModelMixin, BaseUuidModel):
    class Meta(DeathReportModelMixin.Meta):
        pass


class DeathReportTmg(DeathReportTmgModelMixin, BaseUuidModel):
    class Meta(DeathReportTmgModelMixin.Meta):
        pass


class DeathReportTmgSecond(DeathReportTmgSecondModelMixin, BaseUuidModel):
    class Meta(DeathReportTmgSecondModelMixin.Meta):
        pass



from django.conf import settings
from django.contrib.admin import display
from django.core.exceptions import ObjectDoesNotExist
from django.urls.base import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from edc_utils.text import convert_php_dateformat

from ..utils import get_ae_model
from .utils import ColumnItem


class NonAeInitialModelAdminMixin:
    add_form_template: str = "edc_adverse_event/admin/change_form.html"
    change_list_template = "edc_adverse_event/admin/change_list.html"
    change_form_template = "edc_adverse_event/admin/change_form.html"

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "ae_initial":
            if request.GET.get("ae_initial"):
                kwargs["queryset"] = get_ae_model("aeinitial").objects.filter(
                    id__exact=request.GET.get("ae_initial", 0)
                )
            else:
                kwargs["queryset"] = get_ae_model("aeinitial").objects.none()
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def get_readonly_fields(self, request, obj=None) -> tuple:
        fields = super().get_readonly_fields(request, obj=obj)
        if obj:
            fields += ("ae_initial",)
        return fields

    def initial_ae(self, obj):
        """Returns a shortened action identifier."""
        if obj.ae_initial:
            url_name = "_".join(obj.ae_initial._meta.label_lower.split("."))
            namespace = self.admin_site.name
            url = reverse(f"{namespace}:{url_name}_changelist")
            return format_html(  # nosec B703, B308
                '<a data-toggle="tooltip" title="go to AE initial" href="{}?q={}">{}</a>',
                mark_safe(url),  # nosec B703, B308
                obj.ae_initial.action_identifier,
                obj.ae_initial.identifier,
            )
        return None


class AdverseEventModelAdminMixin:
    add_form_template: str = "edc_adverse_event/admin/change_form.html"
    change_list_template = "edc_adverse_event/admin/change_list.html"
    change_form_template = "edc_adverse_event/admin/change_form.html"

    @display(description="subject identifier", ordering="subject_identifier")
    def subject_identifier_column(self, obj=None):
        return format_html(f"{obj.subject_identifier}<BR>{obj.action_identifier.upper()[-9:]}")

    @display(description="User", ordering="user_created")
    def user(self, obj=None):
        """Returns formatted usernames and creation/modification dates."""
        return format_html(
            "<BR>".join(
                [
                    obj.user_created,
                    obj.created.strftime(convert_php_dateformat(settings.SHORT_DATE_FORMAT)),
                    obj.user_modified,
                    obj.modified.strftime(convert_php_dateformat(settings.SHORT_DATE_FORMAT)),
                ]
            )
        )

    @display(description="Documents")
    def documents_column(self, model_obj):
        """Returns a formatted list of links to AE Follow up reports."""
        column_items: list[ColumnItem] = []
        ae_followup_model_cls = get_ae_model("aefollowup")
        ae_susar_model_cls = get_ae_model("aesusar")
        death_report_model_cls = get_ae_model("deathreport")

        try:
            ae_initial = model_obj.ae_initial
        except AttributeError:
            ae_initial = model_obj
        column_items.append(
            ColumnItem(self, ae_initial, ae_initial.action_identifier, "ae_start_date")
        )

        try:
            death_report = death_report_model_cls.objects.get(
                subject_identifier=ae_initial.subject_identifier
            )
        except ObjectDoesNotExist:
            pass
        else:
            column_items.append(
                ColumnItem(
                    self, death_report, death_report.subject_identifier, "death_datetime"
                )
            )
        for ae_followup in ae_followup_model_cls.objects.filter(
            related_action_item=ae_initial.action_item
        ):
            column_items.append(
                ColumnItem(
                    self,
                    ae_followup,
                    ae_followup.ae_initial.action_identifier,
                    "outcome_date",
                )
            )
        for ae_susar in ae_susar_model_cls.objects.filter(
            related_action_item=ae_initial.action_item
        ):
            column_items.append(
                ColumnItem(self, ae_susar, ae_susar.ae_initial.action_identifier)
            )
        # order
        sorted(column_items, reverse=True)
        # render
        html = "<table><tr><td>"
        html += "</td><tr><td>".join([c.anchor for c in column_items])
        html += "</td></td></table>"
        return format_html(html)



from __future__ import annotations

from typing import TYPE_CHECKING, Type

from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.auth import get_permission_codename
from django.contrib.messages import ERROR
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext as _
from edc_model_admin.utils import add_to_messages_once
from edc_utils import convert_php_dateformat

from .constants import TMG_ROLE

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest

    from edc_adverse_event.model_mixins import (
        AeFollowupModelMixin,
        AeInitialModelMixin,
        DeathReportModelMixin,
    )


def validate_ae_initial_outcome_date(form_obj):
    ae_initial = form_obj.cleaned_data.get("ae_initial")
    if not ae_initial and form_obj.instance:
        ae_initial = form_obj.instance.ae_initial
    outcome_date = form_obj.cleaned_data.get("outcome_date")
    if ae_initial and outcome_date:
        if outcome_date < ae_initial.ae_start_date:
            formatted_dte = ae_initial.ae_start_date.strftime(
                convert_php_dateformat(settings.SHORT_DATE_FORMAT)
            )
            raise forms.ValidationError(
                {"outcome_date": f"May not be before the AE start date {formatted_dte}."}
            )


def get_adverse_event_admin_site() -> str:
    return getattr(
        settings, "ADVERSE_EVENT_ADMIN_SITE", f"{get_adverse_event_app_label()}_admin"
    )


def get_adverse_event_app_label() -> str:
    app_label = getattr(settings, "ADVERSE_EVENT_APP_LABEL", None)
    if not app_label:
        app_label = getattr(settings, "EDC_ADVERSE_EVENT_APP_LABEL", None)
    if not app_label:
        raise ValueError(
            "Attribute not set. See `get_adverse_event_app_label()` or "
            "`settings.EDC_ADVERSE_EVENT_APP_LABEL`."
        )
    return app_label


def get_hospitalization_model_app_label() -> str:
    return getattr(
        settings, "EDC_ADVERSE_EVENT_HOSPITALIZATION_APP_LABEL", get_adverse_event_app_label()
    )


def get_ae_model(
    model_name,
) -> Type[DeathReportModelMixin] | Type[AeInitialModelMixin] | Type[AeFollowupModelMixin]:
    return django_apps.get_model(f"{get_adverse_event_app_label()}.{model_name}")


def get_ae_model_name(model_name: str) -> str:
    return f"{get_adverse_event_app_label()}.{model_name}"


def has_valid_tmg_perms(request: WSGIRequest, add_message: bool = None):
    """Checks if user has TMG_ROLE but not granted add/change
    perms to any non-TMG AE models.

    add_message: if True, adds a message to message context.
    """
    non_tmg_ae_models = ["aeinitial", "aefollowup", "deathreport"]
    # check role
    try:
        has_valid_perms = request.user.userprofile.roles.get(name=TMG_ROLE)
    except ObjectDoesNotExist:
        has_valid_perms = False
        if add_message:
            add_to_messages_once(
                request,
                ERROR,
                (
                    "Access disabled. User has not been granted a TMG role. "
                    "Contact your administrator."
                ),
            )
    # check AE model perms
    if has_valid_tmg_perms:
        codenames = {}
        for model_name in non_tmg_ae_models:
            model_cls = get_ae_model(model_name)
            codename = get_permission_codename("change", model_cls._meta)
            codenames.update({model_cls: f"{model_cls._meta.app_label}.{codename}"})
        for model_cls, codename in codenames.items():
            if request.user.has_perm(codename):
                if add_message:
                    add_to_messages_once(
                        request,
                        ERROR,
                        (
                            _(
                                "Access disabled. A TMG user may not have change "
                                "permission for any adverse event form. Contact your "
                                "administrator. Got %(verbose_name)s"
                            )
                            % {"verbose_name": model_cls._meta.verbose_name}
                        ),
                    )
                has_valid_perms = False
                break
    return has_valid_perms



from datetime import timedelta

from django import forms
from django.test import TestCase
from edc_constants.constants import NO, NOT_APPLICABLE, UNKNOWN, YES
from edc_form_validators import FormValidatorTestCaseMixin
from edc_form_validators.tests.mixins import FormValidatorTestMixin
from edc_utils import get_utcnow, get_utcnow_as_date

from edc_adverse_event.form_validators import HospitalizationFormValidator as Base


class HospitalizationFormValidator(FormValidatorTestMixin, Base):
    pass


class TestHospitalizationFormValidation(FormValidatorTestCaseMixin, TestCase):
    form_validator_cls = HospitalizationFormValidator

    @staticmethod
    def get_cleaned_data() -> dict:
        return {
            "report_datetime": get_utcnow(),
            "have_details": YES,
            "admitted_date": get_utcnow_as_date() - timedelta(days=3),
            "admitted_date_estimated": NO,
            "discharged": YES,
            "discharged_date": get_utcnow_as_date() - timedelta(days=1),
            "discharged_date_estimated": NO,
            "narrative": "Details of admission",
        }

    def test_cleaned_data_ok(self):
        cleaned_data = self.get_cleaned_data()
        form_validator = HospitalizationFormValidator(cleaned_data=cleaned_data)
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_minimal_details_ok(self):
        cleaned_data = {
            "report_datetime": get_utcnow(),
            "have_details": NO,
            "admitted_date": get_utcnow_as_date(),
            "admitted_date_estimated": NO,
            "discharged": UNKNOWN,
            "discharged_date": None,
            "discharged_date_estimated": NOT_APPLICABLE,
            "narrative": "",
        }
        form_validator = HospitalizationFormValidator(cleaned_data=cleaned_data)
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_discharged_date_required_if_discharged_yes(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "have_details": YES,
                "admitted_date": get_utcnow_as_date(),
                "admitted_date_estimated": NO,
                "discharged": YES,
                "discharged_date": None,
            }
        )
        form_validator = HospitalizationFormValidator(cleaned_data=cleaned_data)
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("discharged_date", cm.exception.error_dict)
        self.assertEqual(
            {"discharged_date": ["This field is required."]},
            cm.exception.message_dict,
        )

    def test_discharged_date_not_required_if_discharged_not_yes(self):
        for answer in [NO, UNKNOWN]:
            with self.subTest(answer=answer):
                cleaned_data = self.get_cleaned_data()
                cleaned_data.update(
                    {
                        "have_details": NO,
                        "admitted_date": get_utcnow_as_date(),
                        "admitted_date_estimated": NO,
                        "discharged": answer,
                        "discharged_date": get_utcnow_as_date(),
                    }
                )
                form_validator = HospitalizationFormValidator(cleaned_data=cleaned_data)
                with self.assertRaises(forms.ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("discharged_date", cm.exception.error_dict)
                self.assertEqual(
                    {"discharged_date": ["This field is not required."]},
                    cm.exception.message_dict,
                )

    def test_discharged_date_after_admitted_date_ok(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "have_details": YES,
                "admitted_date": get_utcnow_as_date() - timedelta(days=3),
                "admitted_date_estimated": NO,
                "discharged": YES,
                "discharged_date": get_utcnow_as_date() - timedelta(days=1),
                "discharged_date_estimated": NO,
            }
        )
        form_validator = HospitalizationFormValidator(cleaned_data=cleaned_data)
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_discharged_date_same_as_admitted_date_ok(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "have_details": YES,
                "admitted_date": get_utcnow_as_date() - timedelta(days=3),
                "admitted_date_estimated": NO,
                "discharged": YES,
                "discharged_date": get_utcnow_as_date() - timedelta(days=3),
                "discharged_date_estimated": NO,
                # CSF date cannot be after date discharged
                "csf_positive_cm_date": get_utcnow_as_date() - timedelta(days=3),
            }
        )
        form_validator = HospitalizationFormValidator(cleaned_data=cleaned_data)
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_discharged_date_raises_if_earlier_than_admitted_date(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "have_details": YES,
                "admitted_date": get_utcnow_as_date() - timedelta(days=3),
                "admitted_date_estimated": NO,
                "discharged": YES,
                "discharged_date": get_utcnow_as_date() - timedelta(days=4),
                "discharged_date_estimated": NO,
            }
        )
        form_validator = HospitalizationFormValidator(cleaned_data=cleaned_data)
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("discharged_date", cm.exception.error_dict)
        self.assertEqual(
            {"discharged_date": ["Invalid. Cannot be before date admitted."]},
            cm.exception.message_dict,
        )

    def test_discharged_date_estimated_applicable_if_discharged_yes(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "have_details": YES,
                "admitted_date": get_utcnow_as_date(),
                "admitted_date_estimated": NO,
                "discharged": YES,
                "discharged_date": get_utcnow_as_date(),
                "discharged_date_estimated": NOT_APPLICABLE,
            }
        )
        form_validator = HospitalizationFormValidator(cleaned_data=cleaned_data)
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("discharged_date_estimated", cm.exception.error_dict)
        self.assertEqual(
            {"discharged_date_estimated": ["This field is applicable."]},
            cm.exception.message_dict,
        )

    def test_discharged_date_estimated_applicable_if_discharged_not_yes(self):
        for answer in [NO, UNKNOWN]:
            with self.subTest(answer=answer):
                cleaned_data = self.get_cleaned_data()
                cleaned_data.update(
                    {
                        "have_details": NO,
                        "admitted_date": get_utcnow_as_date(),
                        "admitted_date_estimated": NO,
                        "discharged": answer,
                        "discharged_date": None,
                        "discharged_date_estimated": "D",
                    }
                )
                form_validator = HospitalizationFormValidator(cleaned_data=cleaned_data)
                with self.assertRaises(forms.ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("discharged_date_estimated", cm.exception.error_dict)
                self.assertEqual(
                    {"discharged_date_estimated": ["This field is not applicable."]},
                    cm.exception.message_dict,
                )

    def test_narrative_required_if_have_details_yes(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "have_details": YES,
                "narrative": "",
            }
        )
        form_validator = HospitalizationFormValidator(cleaned_data=cleaned_data)
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("narrative", cm.exception.error_dict)
        self.assertEqual(
            {"narrative": ["This field is required."]},
            cm.exception.message_dict,
        )

    def test_narrative_can_still_be_entered_if_have_details_no(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "have_details": NO,
                "narrative": "bbbb",
            }
        )
        form_validator = HospitalizationFormValidator(cleaned_data=cleaned_data)
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")



from edc_action_item.action_with_notification import ActionWithNotification
from edc_constants.constants import HIGH_PRIORITY

from ..constants import (
    ADVERSE_EVENT_ADMIN_SITE,
    ADVERSE_EVENT_APP_LABEL,
    AE_INITIAL_ACTION,
    AE_SUSAR_ACTION,
)


class AeSusarAction(ActionWithNotification):
    name = AE_SUSAR_ACTION
    display_name = "Submit AE SUSAR Report"
    notification_display_name = "AE SUSAR Report"
    parent_action_names = [AE_INITIAL_ACTION]
    reference_model = f"{ADVERSE_EVENT_APP_LABEL}.aesusar"
    related_reference_model = f"{ADVERSE_EVENT_APP_LABEL}.aeinitial"
    related_reference_fk_attr = "ae_initial"
    create_by_user = False
    show_link_to_changelist = True
    admin_site_name = ADVERSE_EVENT_ADMIN_SITE
    instructions = "Complete the AE SUSAR report"
    priority = HIGH_PRIORITY



from unittest.mock import PropertyMock, patch

from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.test import TestCase, override_settings
from edc_action_item.get_action_type import get_action_type
from edc_action_item.models.action_item import ActionItem
from edc_constants.constants import CLOSED, DEAD, NEW, NO, YES
from edc_constants.disease_constants import ANAEMIA
from edc_list_data.site_list_data import site_list_data
from edc_ltfu.constants import LOST_TO_FOLLOWUP
from edc_registration.models import RegisteredSubject
from edc_registration.utils import RegisteredSubjectDoesNotExist
from edc_reportable import GRADE5
from edc_utils import get_utcnow
from edc_visit_schedule.utils import OnScheduleError
from model_bakery import baker

from adverse_event_app import list_data
from adverse_event_app.action_items import (
    AeFollowupAction,
    AeInitialAction,
    StudyTerminationConclusionAction,
)
from adverse_event_app.models import AeFollowup, AeInitial, AeSusar, AeTmg
from edc_adverse_event.models import AeClassification

from ...constants import CONTINUING_UPDATE, RECOVERED, RECOVERING


@override_settings(EDC_LIST_DATA_ENABLE_AUTODISCOVER=False)
class TestAeAndActions(TestCase):
    @classmethod
    def setUpClass(cls):
        site_list_data.initialize()
        site_list_data.register(list_data, app_name="adverse_event_app")
        site_list_data.load_data()
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        self.subject_identifier = "12345"
        RegisteredSubject.objects.create(subject_identifier=self.subject_identifier)

    def tearDown(self):
        RegisteredSubject.objects.all().delete()

    def test_subject_identifier(self):
        baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )

        self.assertRaises(
            RegisteredSubjectDoesNotExist,
            baker.make_recipe,
            "adverse_event_app.aeinitial",
            subject_identifier="blahblah",
        )

    def test_entire_flow(self):
        for index in range(0, 5):
            subject_identifier = f"ABCDEF-{index}"
            RegisteredSubject.objects.create(subject_identifier=subject_identifier)
            ae_initial = baker.make_recipe(
                "adverse_event_app.aeinitial", subject_identifier=subject_identifier
            )
            baker.make_recipe(
                "adverse_event_app.aefollowup",
                ae_initial=ae_initial,
                subject_identifier=subject_identifier,
                outcome=RECOVERING,
            )
            baker.make_recipe(
                "adverse_event_app.aefollowup",
                ae_initial=ae_initial,
                subject_identifier=subject_identifier,
                outcome=RECOVERING,
            )
            baker.make_recipe(
                "adverse_event_app.aefollowup",
                ae_initial=ae_initial,
                subject_identifier=subject_identifier,
                outcome=RECOVERING,
            )
            baker.make_recipe(
                "adverse_event_app.aefollowup",
                ae_initial=ae_initial,
                subject_identifier=subject_identifier,
                outcome=RECOVERED,
                followup=NO,
            )

    def test_fk1(self):
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )
        baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERING,
        )
        baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERING,
        )
        baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERING,
        )
        baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERING,
        )

    def test_fk2(self):
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )
        baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERING,
        )
        baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERING,
        )
        baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERED,
            followup=NO,
        )

    def test_ae_initial_action(self):
        """Asserts an AeInitial creates one and only one
        AeFollowupAction.
        """
        # create ae initial action
        action_type = get_action_type(AeInitialAction)
        action_item = ActionItem.objects.create(
            subject_identifier=self.subject_identifier, action_type=action_type
        )
        # create ae initial
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial",
            action_identifier=action_item.action_identifier,
            subject_identifier=self.subject_identifier,
        )
        ActionItem.objects.get(
            subject_identifier=self.subject_identifier, action_type=action_type
        )
        action_item = ActionItem.objects.get(
            subject_identifier=self.subject_identifier,
            action_type=action_type,
            status=CLOSED,
        )

        # assert ae initial action created ONE ae followup
        action_type = get_action_type(AeFollowupAction)
        self.assertEqual(
            ActionItem.objects.filter(
                subject_identifier=self.subject_identifier, action_type=action_type
            ).count(),
            1,
        )

        # assert ae initial action created ONE ae followup
        # with correct parent action identifier
        ActionItem.objects.get(
            subject_identifier=self.subject_identifier,
            parent_action_item=action_item,
            action_type=action_type,
            status=NEW,
        )

        # resave ae initial and show does not create another followup
        ae_initial.save()
        ae_initial.save()
        ae_initial.save()
        self.assertEqual(
            ActionItem.objects.filter(
                subject_identifier=self.subject_identifier, action_type=action_type
            ).count(),
            1,
        )

    def test_ae_initial_action2(self):
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )
        action_item = ActionItem.objects.get(
            subject_identifier=self.subject_identifier,
            action_identifier=ae_initial.action_identifier,
            reference_model="adverse_event_app.aeinitial",
        )
        self.assertEqual(action_item.status, CLOSED)

    def test_ae_initial_creates_action(self):
        # create reference model first which creates action_item
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )
        try:
            ActionItem.objects.get(
                subject_identifier=self.subject_identifier,
                action_identifier=ae_initial.action_identifier,
                reference_model="adverse_event_app.aeinitial",
            )
        except ObjectDoesNotExist:
            self.fail("action item unexpectedly does not exist.")
        except MultipleObjectsReturned:
            self.fail("action item unexpectedly returned multiple objects.")
        self.assertEqual(
            ActionItem.objects.filter(
                subject_identifier=self.subject_identifier,
                action_identifier=ae_initial.action_identifier,
                reference_model="adverse_event_app.aeinitial",
            ).count(),
            1,
        )
        self.assertEqual(
            ActionItem.objects.filter(
                subject_identifier=self.subject_identifier,
                parent_action_item=ae_initial.action_item,
                related_action_item=ae_initial.action_item,
                reference_model="adverse_event_app.aefollowup",
            ).count(),
            1,
        )

    def test_ae_initial_does_not_recreate_action_on_resave(self):
        # create reference model first which creates action_item
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )
        ae_initial = AeInitial.objects.get(pk=ae_initial.pk)
        ae_initial.save()
        self.assertEqual(
            ActionItem.objects.filter(
                subject_identifier=self.subject_identifier,
                action_identifier=ae_initial.action_identifier,
                reference_model="adverse_event_app.aeinitial",
            ).count(),
            1,
        )

    def test_ae_initial_updates_existing_action_item(self):
        action_type = get_action_type(AeInitialAction)
        action_item = ActionItem.objects.create(
            subject_identifier=self.subject_identifier,
            action_type=action_type,
            reference_model="adverse_event_app.aeinitial",
        )

        # then create reference model
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            action_identifier=action_item.action_identifier,
        )

        action_item = ActionItem.objects.get(pk=action_item.pk)
        self.assertEqual(action_item.reference_model, ae_initial._meta.label_lower)
        self.assertEqual(action_item.action_identifier, ae_initial.action_identifier)

    def test_ae_initial_creates_next_action_on_close(self):
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )
        ae_initial = AeInitial.objects.get(pk=ae_initial.pk)
        self.assertTrue(
            ActionItem.objects.get(
                subject_identifier=self.subject_identifier,
                action_identifier=ae_initial.action_identifier,
                parent_action_item=None,
                reference_model="adverse_event_app.aeinitial",
                status=CLOSED,
            )
        )
        self.assertTrue(
            ActionItem.objects.get(
                subject_identifier=self.subject_identifier,
                parent_action_item=ae_initial.action_item,
                related_action_item=ae_initial.action_item,
                reference_model="adverse_event_app.aefollowup",
                status=NEW,
            )
        )

    def test_next_action1(self):
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )
        # action item has no parent, is updated
        ActionItem.objects.get(
            parent_action_item=None,
            action_identifier=ae_initial.action_identifier,
            reference_model="adverse_event_app.aeinitial",
        )

        # action item a parent, is not updated
        ActionItem.objects.get(
            parent_action_item=ae_initial.action_item,
            related_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.aefollowup",
        )

    def test_next_action2(self):
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )
        ae_followup = baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERING,
        )
        ae_followup = AeFollowup.objects.get(pk=ae_followup.pk)

        ActionItem.objects.get(
            parent_action_item=ae_initial.action_item,
            related_action_item=ae_initial.action_item,
            action_identifier=ae_followup.action_identifier,
            reference_model="adverse_event_app.aefollowup",
            linked_to_reference=True,
            status=CLOSED,
        )
        ActionItem.objects.get(
            parent_action_item=ae_followup.action_item,
            related_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.aefollowup",
            linked_to_reference=False,
            status=NEW,
        )

    def test_next_action3(self):
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )
        ae_followup1 = baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERING,
        )
        ae_followup1 = AeFollowup.objects.get(pk=ae_followup1.pk)
        ae_followup2 = baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERING,
        )
        ae_followup2 = AeFollowup.objects.get(pk=ae_followup2.pk)
        ActionItem.objects.get(
            parent_action_item=ae_initial.action_item,
            related_action_item=ae_initial.action_item,
            action_identifier=ae_followup1.action_identifier,
            reference_model="adverse_event_app.aefollowup",
            linked_to_reference=True,
            status=CLOSED,
        )
        ActionItem.objects.get(
            parent_action_item=ae_followup1.action_item,
            related_action_item=ae_initial.action_item,
            action_identifier=ae_followup2.action_identifier,
            reference_model="adverse_event_app.aefollowup",
            linked_to_reference=True,
            status=CLOSED,
        )
        ActionItem.objects.get(
            parent_action_item=ae_followup2.action_item,
            related_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.aefollowup",
            linked_to_reference=False,
            status=NEW,
        )

    def test_next_action4(self):
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )
        ae_followup1 = baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=CONTINUING_UPDATE,
            followup=YES,
        )
        ae_followup1 = AeFollowup.objects.get(pk=ae_followup1.pk)
        # set followup = NO so next action item is not created
        ae_followup2 = baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERED,
            followup=NO,
        )
        ae_followup2 = AeFollowup.objects.get(pk=ae_followup2.pk)

        ActionItem.objects.get(
            parent_action_item=ae_initial.action_item,
            related_action_item=ae_initial.action_item,
            action_identifier=ae_followup1.action_identifier,
            reference_model="adverse_event_app.aefollowup",
            linked_to_reference=True,
            status=CLOSED,
        )

        ActionItem.objects.get(
            parent_action_item=ae_followup1.action_item,
            related_action_item=ae_initial.action_item,
            action_identifier=ae_followup2.action_identifier,
            reference_model="adverse_event_app.aefollowup",
            linked_to_reference=True,
            status=CLOSED,
        )

        self.assertRaises(
            ObjectDoesNotExist,
            ActionItem.objects.get,
            parent_action_item=ae_followup2.action_item,
            related_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.aefollowup",
            linked_to_reference=False,
            status=NEW,
        )

    def test_next_action5(self):
        anaemia = AeClassification.objects.get(name=ANAEMIA)
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            ae_classification=anaemia,
        )
        ae_initial = AeInitial.objects.get(pk=ae_initial.pk)

        ActionItem.objects.get(
            parent_action_item=None,
            related_action_item=None,
            action_identifier=ae_initial.action_identifier,
            reference_model="adverse_event_app.aeinitial",
            linked_to_reference=True,
            status=CLOSED,
        )

        ActionItem.objects.get(
            parent_action_item=ae_initial.action_item,
            related_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.aetmg",
            linked_to_reference=False,
            status=NEW,
        )

        # note: ae_classification matches ae_initial
        ae_tmg = baker.make_recipe(
            "adverse_event_app.aetmg",
            subject_identifier=self.subject_identifier,
            ae_initial=ae_initial,
            ae_classification=anaemia.name,
            report_status=CLOSED,
        )

        ae_tmg = AeTmg.objects.get(pk=ae_tmg.pk)

        ActionItem.objects.get(
            parent_action_item=ae_initial.action_item,
            related_action_item=ae_initial.action_item,
            action_identifier=ae_tmg.action_identifier,
            reference_model="adverse_event_app.aetmg",
            linked_to_reference=True,
            status=CLOSED,
        )

    def test_ae_followup_multiple_instances(self):
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )
        ae_initial = AeInitial.objects.get(pk=ae_initial.pk)

        ae_followup = baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERING,
        )
        AeFollowup.objects.get(pk=ae_followup.pk)

        ae_followup = baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            outcome=RECOVERING,
        )
        AeFollowup.objects.get(pk=ae_followup.pk)

    @patch("edc_adverse_event.action_items.ae_followup_action.site_action_items.get_by_model")
    @patch.object(AeFollowupAction, "offschedule_models", new_callable=PropertyMock)
    @patch.object(AeFollowupAction, "onschedule_models", new_callable=PropertyMock)
    def test_ae_followup_outcome_ltfu_creates_action(
        self, mock_onschedule_models, mock_offschedule_models, mock_get_by_model
    ):
        mock_onschedule_models.return_value = ["adverse_event_app.subjectconsent"]
        mock_offschedule_models.return_value = ["adverse_event_app.studyterminationconclusion"]
        mock_get_by_model.return_value = StudyTerminationConclusionAction

        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )

        ae_followup = baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            report_datetime=get_utcnow(),
            outcome=LOST_TO_FOLLOWUP,
        )
        try:
            ActionItem.objects.get(
                parent_action_item=ae_followup.action_item,
                reference_model="adverse_event_app.studyterminationconclusion",
            )
        except ObjectDoesNotExist:
            self.fail("ObjectDoesNotExist unexpectedly raised")

    @patch("edc_adverse_event.action_items.ae_followup_action.site_action_items.get_by_model")
    @patch.object(AeFollowupAction, "offschedule_models", new_callable=PropertyMock)
    @patch.object(AeFollowupAction, "onschedule_models", new_callable=PropertyMock)
    def test_ae_followup_outcome_ltfu_raises(
        self, mock_onschedule_models, mock_offschedule_models, mock_get_by_model
    ):
        mock_onschedule_models.return_value = []  # not on schedule
        mock_offschedule_models.return_value = []
        mock_get_by_model.return_value = StudyTerminationConclusionAction

        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )

        self.assertRaises(
            OnScheduleError,
            baker.make_recipe,
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            report_datetime=get_utcnow(),
            outcome=LOST_TO_FOLLOWUP,
        )

    @patch("edc_adverse_event.action_items.ae_followup_action.site_action_items.get_by_model")
    @patch.object(AeFollowupAction, "offschedule_models", new_callable=PropertyMock)
    def test_ae_followup_outcome_not_ltfu(self, mock_offschedule_models, mock_get_by_model):
        mock_offschedule_models.return_value = ["adverse_event_app.studyterminationconclusion"]
        mock_get_by_model.return_value = StudyTerminationConclusionAction

        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial", subject_identifier=self.subject_identifier
        )

        ae_followup = baker.make_recipe(
            "adverse_event_app.aefollowup",
            ae_initial=ae_initial,
            subject_identifier=self.subject_identifier,
            report_datetime=get_utcnow(),
            outcome=DEAD,
        )

        try:
            ActionItem.objects.get(
                parent_action_item=ae_followup.action_item,
                reference_model="adverse_event_app.studyterminationconclusion",
            )
        except ObjectDoesNotExist:
            pass
        else:
            self.fail("ObjectDoesNotExist unexpectedly raised")

    def test_ae_creates_death_report_action(self):
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            ae_grade=GRADE5,
            sae=NO,
        )

        ActionItem.objects.get(
            parent_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.deathreport",
        )

        ActionItem.objects.get(
            parent_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.aetmg",
        )

    def test_ae_initial_creates_susar_if_not_reported(self):
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            susar=YES,
            susar_reported=YES,
            user_created="erikvw",
        )

        self.assertRaises(
            ObjectDoesNotExist,
            ActionItem.objects.get,
            parent_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.aesusar",
        )

        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            susar=YES,
            susar_reported=NO,
            user_created="erikvw",
        )

        ActionItem.objects.get(
            parent_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.aesusar",
        )

    def test_susar_updates_aeinitial_if_submitted(self):
        # create ae initial
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            susar=YES,
            susar_reported=NO,
            user_created="erikvw",
        )

        # confirm ae susar action item is created
        action_item = ActionItem.objects.get(
            parent_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.aesusar",
        )

        self.assertEqual(action_item.status, NEW)

        # create ae susar
        baker.make_recipe(
            "adverse_event_app.aesusar",
            subject_identifier=self.subject_identifier,
            submitted_datetime=get_utcnow(),
            ae_initial=ae_initial,
        )

        # confirm action status is closed
        action_item.refresh_from_db()
        self.assertEqual(action_item.status, CLOSED)

        # confirm susar updates ae_initial (through signal)
        ae_initial.refresh_from_db()
        self.assertEqual(ae_initial.susar_reported, YES)

    def test_aeinitial_can_close_action_without_susar_model(self):
        # create ae initial
        ae_initial = baker.make_recipe(
            "adverse_event_app.aeinitial",
            subject_identifier=self.subject_identifier,
            susar=YES,
            susar_reported=NO,
            user_created="erikvw",
        )

        # confirm ae susar action item is created
        action_item = ActionItem.objects.get(
            parent_action_item=ae_initial.action_item,
            reference_model="adverse_event_app.aesusar",
        )

        # change to YES before submitting an AeSusar
        ae_initial.susar_reported = YES
        ae_initial.save()
        ae_initial.refresh_from_db()

        # confirm AeSusar was created (by signal)
        try:
            AeSusar.objects.get(ae_initial=ae_initial)
        except ObjectDoesNotExist:
            self.fail("AeSusar unexpectedly does not exist")

        # confirm ActionItem is closed
        action_item.refresh_from_db()
        self.assertEqual(action_item.status, CLOSED)



from edc_dashboard.utils import insert_bootstrap_version

dashboard_templates = dict(
    ae_home_template="edc_adverse_event/ae/ae_home.html",
    ae_listboard_template="edc_adverse_event/ae/ae_listboard.html",
    ae_death_report_listboard_template="edc_adverse_event/ae/ae_death_report_listboard.html",
    tmg_ae_listboard_template="edc_adverse_event/tmg/tmg_ae_listboard.html",
    tmg_ae_listboard_result_template="edc_adverse_event/tmg/tmg_ae_listboard_result.html",
    tmg_death_listboard_template="edc_adverse_event/tmg/tmg_death_listboard.html",
    tmg_home_template="edc_adverse_event/tmg/tmg_home.html",
    tmg_summary_listboard_template="edc_adverse_event/tmg/tmg_summary_listboard.html",
)
dashboard_templates = insert_bootstrap_version(**dashboard_templates)



from __future__ import annotations

from datetime import date

from django.core.exceptions import ObjectDoesNotExist
from edc_constants.constants import CLOSED, NO, OTHER
from edc_form_validators import INVALID_ERROR, FormValidator

from ..form_validator_mixins import (
    BaseRequiresDeathReportFormValidatorMixin,
    DeathReportFormValidatorMixin,
)
from ..utils import get_ae_model


class DeathReportTmgFormValidator(
    DeathReportFormValidatorMixin, BaseRequiresDeathReportFormValidatorMixin, FormValidator
):
    @property
    def death_date(self) -> date:
        try:
            obj = get_ae_model("deathreport").objects.get(
                subject_identifier=self.cleaned_data.get("subject_identifier")
            )
        except ObjectDoesNotExist:
            self.raise_validation_error("Death report not found.", INVALID_ERROR)
        death_date = getattr(obj, getattr(obj, "death_date_field"))
        try:
            death_date = death_date.date()
        except AttributeError:
            pass
        return death_date

    def clean(self):
        self.date_is_after_or_raise(
            field="report_datetime",
            reference_value=self.death_report_date,
            inclusive=True,
        )

        self.required_if(
            CLOSED,
            field="report_status",
            field_required="cause_of_death",
            inverse=False,
        )

        self.validate_other_specify(
            field="cause_of_death",
            other_specify_field="cause_of_death_other",
            other_stored_value=OTHER,
        )

        self.required_if(
            CLOSED,
            field="report_status",
            field_required="cause_of_death_agreed",
            inverse=False,
        )

        self.required_if(
            NO, field="cause_of_death_agreed", field_required="narrative", inverse=False
        )

        self.required_if(
            CLOSED, field="report_status", field_required="report_closed_datetime"
        )

        if self.cleaned_data.get("report_closed_datetime"):
            self.date_is_after_or_raise(
                field="report_closed_datetime",
                reference_field="report_datetime",
                inclusive=True,
                msg="Must be on or after the report date above.",
            )



from django import forms

from ..modelform_mixins import AeInitialModelFormMixin, AeModelFormMixin
from ..utils import get_ae_model


class AeInitialForm(AeInitialModelFormMixin, AeModelFormMixin, forms.ModelForm):
    class Meta(AeInitialModelFormMixin.Meta):
        model = get_ae_model("aeinitial")
        fields = "__all__"



from datetime import date

from dateutil.relativedelta import relativedelta
from django import forms
from django.test import TestCase
from edc_facility import import_holidays
from edc_form_validators import FormValidator

from adverse_event_app.models import DeathReportTmg
from edc_adverse_event.form_validator_mixins import (
    RequiresDeathReportFormValidatorMixin,
)

from .mixins import DeathReportTestMixin


class TestFormValidators(DeathReportTestMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        import_holidays()
        super().setUpClass()

    def test_death_report_not_found(self):
        class TestFormValidator(RequiresDeathReportFormValidatorMixin, FormValidator):
            death_date_field = "death_date"

            def clean(self):
                self.match_date_of_death_or_raise()

        data = dict(subject_identifier=self.subject_identifier, death_date=date(2000, 1, 1))
        form_validator = TestFormValidator(cleaned_data=data, model=DeathReportTmg)
        self.assertRaises(forms.ValidationError, form_validator.validate)
        self.assertIn("not found", str(form_validator._errors.get("__all__")))

    def test_death_report_found(self):
        class TestFormValidator(RequiresDeathReportFormValidatorMixin, FormValidator):
            death_date_field = "death_date"

            def clean(self):
                self.match_date_of_death_or_raise()

        death_report = self.get_death_report()
        data = dict(
            subject_identifier=self.subject_identifier,
            death_date=death_report.death_datetime.date(),
        )
        form_validator = TestFormValidator(cleaned_data=data, model=DeathReportTmg)

        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_death_report_date(self):
        class TestFormValidator(RequiresDeathReportFormValidatorMixin, FormValidator):
            death_date_field = "death_date"

            def clean(self):
                self.match_date_of_death_or_raise()

        data = dict(subject_identifier=self.subject_identifier, death_date=None)
        form_validator = TestFormValidator(cleaned_data=data, model=DeathReportTmg)
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

        # specify date of death when there is not death report, raises
        data = dict(subject_identifier=self.subject_identifier, death_date=date(2000, 1, 1))
        form_validator = TestFormValidator(cleaned_data=data, model=DeathReportTmg)
        self.assertRaises(forms.ValidationError, form_validator.validate)
        self.assertIn("not found", str(form_validator._errors.get("__all__")))

        # add a death report
        death_report = self.get_death_report()
        # use wrong date of death, raises
        data = dict(
            subject_identifier=self.subject_identifier,
            death_date=death_report.death_datetime.date() - relativedelta(days=1),
        )
        form_validator = TestFormValidator(cleaned_data=data, model=DeathReportTmg)
        self.assertRaises(forms.ValidationError, form_validator.validate)
        self.assertIn("Date does not match", str(form_validator._errors.get("death_date")))

        # use correct date of death, ok
        data = dict(
            subject_identifier=self.subject_identifier,
            death_date=death_report.death_datetime.date(),
        )
        form_validator = TestFormValidator(cleaned_data=data, model=DeathReportTmg)
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

        # use correct date of death as datetime, ok
        data = dict(
            subject_identifier=self.subject_identifier,
            death_date=death_report.death_datetime,
        )
        death_report.death_datetime = death_report.death_datetime.date()
        death_report.save()
        form_validator = TestFormValidator(cleaned_data=data, model=DeathReportTmg)
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_death_report_date2(self):
        class TestFormValidator(RequiresDeathReportFormValidatorMixin, FormValidator):
            death_date_field = "death_date"

            def clean(self):
                self.match_date_of_death_or_raise()

        # add a death report
        death_report = self.get_death_report()
        death_report.death_date = death_report.death_datetime.date()
        death_report.death_datetime = None
        death_report.save()
        death_report.refresh_from_db()

        data = dict(
            subject_identifier=self.subject_identifier,
            death_date=death_report.death_date,
        )

        form_validator = TestFormValidator(cleaned_data=data, model=DeathReportTmg)
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")



from edc_constants.constants import YES
from edc_form_validators import FormValidator


class AeInitialFormValidator(FormValidator):
    def clean(self):
        self.validate_other_specify(field="ae_classification")

        self.validate_relationship_to_study_drug()

        self.required_if(YES, field="ae_cause", field_required="ae_cause_other")

        self.applicable_if(YES, field="sae", field_applicable="sae_reason")

        self.applicable_if(YES, field="susar", field_applicable="susar_reported")

    def validate_relationship_to_study_drug(self):
        pass



from __future__ import annotations

import os
from textwrap import wrap
from typing import TYPE_CHECKING

from django import template
from django.conf import settings
from django.contrib.messages import ERROR
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import select_template
from django.utils.html import format_html
from django.utils.translation import gettext as _
from edc_action_item.utils import get_reference_obj
from edc_constants.constants import CLOSED, OPEN, OTHER, YES
from edc_dashboard.utils import get_bootstrap_version
from edc_model_admin.utils import add_to_messages_once
from edc_utils import get_utcnow

from ..constants import AE_WITHDRAWN, TMG_ROLE
from ..utils import get_adverse_event_app_label, get_ae_model, has_valid_tmg_perms
from ..view_utils import TmgButton

if TYPE_CHECKING:
    from django.db.models import QuerySet
    from edc_action_item.models import ActionItem
    from edc_model.models import BaseUuidModel

    from ..model_mixins import (
        AeFollowupModelMixin,
        AeInitialModelMixin,
        DeathReportModelMixin,
        DeathReportTmgModelMixin,
    )

    class DeathReportTmgModel(DeathReportTmgModelMixin, BaseUuidModel):
        ...

    class DeathReportTmgSecondModel(DeathReportTmgModelMixin, BaseUuidModel):
        ...

    class AeInitialModel(AeInitialModelMixin, BaseUuidModel):
        ...

    class AeFollowupModel(AeFollowupModelMixin, BaseUuidModel):
        ...

    class DeathReportModel(DeathReportModelMixin, BaseUuidModel):
        ...


register = template.Library()


def wrapx(text, length):
    if length:
        return "<BR>".join(wrap(text, length))
    return text


def select_ae_template(relative_path):
    """Returns a template object."""
    local_path = f"{get_adverse_event_app_label()}/bootstrap{get_bootstrap_version()}/"
    default_path = f"edc_adverse_event/bootstrap{get_bootstrap_version()}/"
    return select_template(
        [
            os.path.join(local_path, relative_path),
            os.path.join(default_path, relative_path),
        ]
    )


def select_description_template(model):
    """Returns a template name."""
    return select_ae_template(f"{model}_description.html").template.name


@register.inclusion_tag(select_description_template("aeinitial"), takes_context=True)
def format_ae_description(context, ae_initial, wrap_length):
    context["utc_date"] = get_utcnow().date()
    context["SHORT_DATE_FORMAT"] = settings.SHORT_DATE_FORMAT
    context["OTHER"] = OTHER
    context["YES"] = YES
    context["ae_initial"] = ae_initial
    try:
        context["sae_reason"] = format_html(wrapx(ae_initial.sae_reason.name, wrap_length))
    except AttributeError:
        context["sae_reason"] = ""
    context["ae_description"] = format_html(wrapx(ae_initial.ae_description, wrap_length))
    return context


@register.inclusion_tag(select_description_template("aefollowup"), takes_context=True)
def format_ae_followup_description(context, ae_followup, wrap_length):
    context["AE_WITHDRAWN"] = AE_WITHDRAWN
    context["utc_date"] = get_utcnow().date()
    context["SHORT_DATE_FORMAT"] = settings.SHORT_DATE_FORMAT
    context["OTHER"] = OTHER
    context["YES"] = YES
    context["ae_followup"] = ae_followup
    context["ae_initial"] = ae_followup.ae_initial
    try:
        context["sae_reason"] = format_html(
            wrapx(ae_followup.ae_initial.sae_reason.name, wrap_length)
        )
    except AttributeError:
        context["sae_reason"] = ""
    context["relevant_history"] = format_html(wrapx(ae_followup.relevant_history, wrap_length))
    context["ae_description"] = format_html(
        wrapx(ae_followup.ae_initial.ae_description, wrap_length)
    )
    return context


@register.inclusion_tag(select_description_template("aesusar"), takes_context=True)
def format_ae_susar_description(context, ae_susar, wrap_length):
    context["utc_date"] = get_utcnow().date()
    context["SHORT_DATE_FORMAT"] = settings.SHORT_DATE_FORMAT
    context["OTHER"] = OTHER
    context["YES"] = YES
    context["ae_susar"] = ae_susar
    context["ae_initial"] = ae_susar.ae_initial
    context["sae_reason"] = format_html(
        "<BR>".join(wrap(ae_susar.ae_initial.sae_reason.name, wrap_length or 35))
    )
    context["ae_description"] = format_html(
        wrapx(ae_susar.ae_initial.ae_description, wrap_length)
    )
    return context


@register.inclusion_tag(
    f"edc_adverse_event/bootstrap{get_bootstrap_version()}/"
    f"tmg/tmg_ae_listboard_result.html",
    takes_context=True,
)
def tmg_listboard_results(
    context, results: [ActionItem], empty_qs_message: str | None = None
) -> dict:
    context["results"] = results
    context["empty_message"] = empty_qs_message
    return context


@register.inclusion_tag(
    f"edc_adverse_event/bootstrap{get_bootstrap_version()}/tmg/death_report_tmg_panel.html",
    takes_context=True,
)
def render_death_report_tmg_panel(context, action_item: ActionItem = None):
    return dict(action_item=action_item)


@register.simple_tag
def death_report_tmg_queryset(action_item: ActionItem = None) -> QuerySet[DeathReportTmgModel]:
    return get_ae_model("deathreporttmg").objects.filter(
        subject_identifier=action_item.subject_identifier
    )


@register.simple_tag
def death_report_tmg2_queryset(
    action_item: ActionItem = None,
) -> QuerySet[DeathReportTmgSecondModel]:
    return get_ae_model("deathreporttmgsecond").objects.filter(
        subject_identifier=action_item.subject_identifier
    )


@register.simple_tag
def death_report_queryset(
    subject_identifier: str = None,
) -> QuerySet[DeathReportTmgSecondModel]:
    return get_ae_model("deathreport").objects.filter(subject_identifier=subject_identifier)


@register.simple_tag
def ae_followup_queryset(
    ae_initial: AeInitialModel = None,
) -> QuerySet[AeFollowupModel] | None:
    if ae_initial:
        return get_ae_model("aefollowup").objects.filter(ae_initial_id=ae_initial.id)
    return None


@register.inclusion_tag(
    f"edc_adverse_event/bootstrap{get_bootstrap_version()}/tmg/ae_tmg_panel.html",
    takes_context=True,
)
def render_tmg_panel(
    context,
    action_item: ActionItem = None,
    reference_obj: DeathReportTmgModel = None,
    view_only: bool | None = None,
    by_user_created_only: bool | None = None,
    counter: int = None,
    report_status: str | None = None,
) -> dict:
    reference_obj = reference_obj or get_reference_obj(action_item)
    if not action_item and reference_obj:
        action_item = reference_obj.action_item
    disable_all = True if not has_valid_tmg_perms(request=context["request"]) else False
    btn = TmgButton(
        user=context["request"].user,
        model_obj=reference_obj,
        model_cls=action_item.action_cls.reference_model_cls(),
        request=context["request"],
        only_user_created_may_access=by_user_created_only,
        forloop_counter=counter,
        current_site=context["request"].site,
        disable_all=disable_all,
    )
    if view_only:
        panel_color = "info"
    elif not reference_obj:
        panel_color = "warning"
    else:
        panel_color = "success"

    # panel_label
    display_name = action_item.display_name.replace("Submit", "").replace("pending", "")
    identifier = action_item.identifier or "New"
    panel_label = _(f"{display_name} {identifier}")

    return dict(
        btn=btn,
        panel_color=panel_color,
        reference_obj=reference_obj,
        action_item=action_item,
        OPEN=OPEN,
        CLOSED=CLOSED,
        report_status=report_status,
        panel_label=panel_label,
    )


@register.simple_tag(takes_context=True)
def has_perms_for_tmg_role(context):
    has_perms = False
    try:
        has_perms = context["request"].user.userprofile.roles.get(name=TMG_ROLE)
    except ObjectDoesNotExist:
        add_to_messages_once(
            context["request"],
            ERROR,
            _(
                "Access disabled. User has not been granted a TMG role. "
                "Contact your administrator."
            ),
        )

    return has_perms


@register.simple_tag()
def get_empty_qs_message(status: str, search_term: str):
    msg = f"There are no {status} reports."
    if search_term:
        msg = f"{msg[:-1]} for your search criteria"
    return _(msg)



from edc_list_data.model_mixins import ListUuidModelMixin


class AeActionClassification(ListUuidModelMixin):
    """Classification of action taken"""

    class Meta(ListUuidModelMixin.Meta):
        pass



from django import forms
from django.test import TestCase

from adverse_event_app.models import AeFollowup
from edc_adverse_event.modelform_mixins import AeFollowupModelFormMixin


class TestModelformMixins(TestCase):
    def test_(self):
        class AeFollowupForm(AeFollowupModelFormMixin, forms.ModelForm):
            class Meta(AeFollowupModelFormMixin.Meta):
                model = AeFollowup
                fields = "__all__"



from django.db import models
from edc_model.validators import datetime_not_future
from edc_utils import get_utcnow

from ...utils import get_adverse_event_app_label


class AeSusarFieldsModelMixin(models.Model):
    ae_initial = models.ForeignKey(
        f"{get_adverse_event_app_label()}.aeinitial", on_delete=models.PROTECT
    )

    report_datetime = models.DateTimeField(
        verbose_name="Report date and time",
        validators=[datetime_not_future],
        default=get_utcnow,
    )

    submitted_datetime = models.DateTimeField(
        verbose_name="AE SUSAR submitted on",
        validators=[datetime_not_future],
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True



from django import forms

from ..form_validators import DeathReportTmgFormValidator
from .ae_modelform_mixin import AeModelFormMixin


class DeathReportTmgSecondModelFormMixin(AeModelFormMixin):
    form_validator_cls = DeathReportTmgFormValidator

    class Meta:
        help_text = {"subject_identifier": "(read-only)", "action_identifier": "(read-only)"}
        widgets = {
            "subject_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
            "action_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



from dateutil.relativedelta import relativedelta
from django.contrib.sites.models import Site
from edc_constants.constants import MALE, NO, NOT_APPLICABLE, YES
from edc_reportable import GRADE4
from edc_utils import get_utcnow
from faker import Faker
from model_bakery.recipe import Recipe, seq

from .models import (
    AeFollowup,
    AeInitial,
    AeSusar,
    AeTmg,
    DeathReport,
    DeathReportTmg,
    DeathReportTmgSecond,
    SubjectConsent,
)

fake = Faker()

subjectconsent = Recipe(
    SubjectConsent,
    consent_datetime=get_utcnow() - relativedelta(months=1),
    dob=get_utcnow() + relativedelta(days=5) - relativedelta(years=25),
    first_name=fake.first_name,
    last_name=fake.last_name,
    initials="AA",
    gender=MALE,
    identity=seq("12315678"),
    confirm_identity=seq("12315678"),
    identity_type="passport",
    is_dob_estimated="-",
    site=Site.objects.get_current(),
)

aeinitial = Recipe(
    AeInitial,
    report_datetime=get_utcnow() - relativedelta(days=5),
    action_identifier=None,
    ae_description="A description of this event",
    ae_grade=GRADE4,
    ae_study_relation_possibility=YES,
    ae_start_date=get_utcnow().date() + relativedelta(days=5),
    ae_awareness_date=get_utcnow().date() + relativedelta(days=5),
    sae=NO,
    susar=NO,
    susar_reported=NOT_APPLICABLE,
    ae_cause=NO,
    ae_cause_other=None,
)

aetmg = Recipe(
    AeTmg,
    action_identifier=None,
    report_datetime=get_utcnow(),
)

aesusar = Recipe(
    AeSusar,
    action_identifier=None,
    report_datetime=get_utcnow(),
)

aefollowup = Recipe(
    AeFollowup,
    relevant_history=NO,
    action_identifier=None,
    report_datetime=get_utcnow(),
)


deathreport = Recipe(
    DeathReport,
    action_identifier=None,
    report_datetime=get_utcnow(),
    death_datetime=get_utcnow(),
)


deathreporttmg = Recipe(
    DeathReportTmg,
    action_identifier=None,
    report_datetime=get_utcnow(),
)


deathreporttmgsecond = Recipe(
    DeathReportTmgSecond,
    action_identifier=None,
    report_datetime=get_utcnow(),
)



from edc_constants.constants import CLOSED, NEW, OPEN

from ...view_mixins import StatusTmgAeListboardView


class NewTmgAeListboardView(StatusTmgAeListboardView):
    listboard_url = "new_tmg_ae_listboard_url"
    search_form_url = "new_tmg_ae_listboard_url"
    status = NEW
    listboard_panel_title = "TMG: New AE Reports"


class OpenTmgAeListboardView(StatusTmgAeListboardView):
    listboard_url = "open_tmg_ae_listboard_url"
    search_form_url = "open_tmg_ae_listboard_url"
    status = OPEN
    listboard_panel_title = "TMG: Open AE Reports"


class ClosedTmgAeListboardView(StatusTmgAeListboardView):
    listboard_url = "closed_tmg_ae_listboard_url"
    search_form_url = "closed_tmg_ae_listboard_url"
    status = CLOSED
    listboard_panel_title = "TMG: Closed AE Reports"



from .ae_followup_fields_model_mixin import AeFollowupFieldsModelMixin
from .ae_followup_methods_model_mixin import AeFollowupMethodsModelMixin
from .ae_followup_model_mixin import AeFollowupModelMixin



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.core.exceptions import ObjectDoesNotExist
from django.utils.html import format_html
from edc_constants.constants import CLOSED, NEW, OPEN
from edc_dashboard.view_mixins import EdcViewMixin
from edc_listboard.view_mixins import ListboardFilterViewMixin, SearchFormViewMixin
from edc_listboard.views import ListboardView as BaseListboardView
from edc_navbar import NavbarViewMixin
from edc_utils import get_utcnow

from ...constants import DEATH_REPORT_ACTION
from ...pdf_reports import DeathPdfReport
from ...utils import get_ae_model

if TYPE_CHECKING:
    from django.db.models import Q


class DeathReportListboardViewMixin(
    NavbarViewMixin,
    EdcViewMixin,
    ListboardFilterViewMixin,
    SearchFormViewMixin,
    BaseListboardView,
):
    listboard_back_url = "ae_home_url"
    home_url = "ae_home_url"
    pdf_report_cls = DeathPdfReport

    listboard_template = "ae_death_report_listboard_template"
    listboard_url = "death_report_listboard_url"
    listboard_panel_style = "default"
    listboard_model = "edc_action_item.actionitem"
    listboard_panel_title = "Adverse Events: Death Reports"
    listboard_view_permission_codename = "edc_adverse_event.view_ae_listboard"
    listboard_instructions = format_html(
        "To find a death report, search on the subject's "
        "study identifier, death report reference number, or AE reference number."
        "<BR>To download the printable report, click on the PDF button"
        "<i class='fas fa-file-pdf fa-fw'></i> "
        "left of the subject's identifier."
    )

    navbar_selected_item = "ae_home"
    ordering = "-report_datetime"
    paginate_by = 25
    search_form_url = "death_report_listboard_url"
    action_type_names = [DEATH_REPORT_ACTION]

    search_fields = [
        "subject_identifier",
        "action_identifier",
        "parent_action_item__action_identifier",
        "related_action_item__action_identifier",
        "user_created",
        "user_modified",
    ]

    def get(self, request, *args, **kwargs):
        if request.GET.get("pdf"):
            response = self.print_pdf_report(
                action_identifier=self.request.GET.get("pdf"), request=request
            )
            return response
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(
            DEATH_REPORT_ACTION=DEATH_REPORT_ACTION,
            utc_date=get_utcnow().date(),
            **self.add_url_to_context(
                new_key="ae_home_url",
                existing_key=self.home_url,
            ),
        )
        return super().get_context_data(**kwargs)

    def get_queryset_filter_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        q_object, options = super().get_queryset_filter_options(request, *args, **kwargs)
        options.update(
            action_type__name__in=self.action_type_names,
            status__in=[NEW, OPEN, CLOSED],
        )
        if kwargs.get("subject_identifier"):
            options.update({"subject_identifier": kwargs.get("subject_identifier")})
        return q_object, options

    @property
    def death_report_model_cls(self):
        return get_ae_model("deathreport")

    def print_pdf_report(self, action_identifier=None, request=None):
        try:
            death_report_obj = self.death_report_model_cls.objects.get(
                action_identifier=action_identifier
            )
        except ObjectDoesNotExist:
            pass
        else:
            pdf_report = self.get_pdf_report(
                pk=death_report_obj.id,
                user=self.request.user,
                request=request,
            )
            return pdf_report.render_to_response()
        return None

    def get_pdf_report(self, **kwargs) -> DeathPdfReport:
        pdf_report_cls = getattr(self.death_report_model_cls, "pdf_report_cls", None)
        if not pdf_report_cls:
            pdf_report_cls = getattr(self, "pdf_report_cls", None)
        return pdf_report_cls(**kwargs)
